@@ -5,6 +5,8 @@
 // <LICENSE-MIT or http://opensource.org/licenses/MIT>, at your
 // option. This file may not be copied, modified, or distributed
 // except according to those terms.
+//
+// SPDX-License-Identifier: MIT/Apache-2.0
 
 use gst::glib;
 use gst::prelude::*;
@@ -12,13 +14,12 @@ use gst::subclass::prelude::*;
 
 use gst::{gst_debug, gst_info};
 use gst_base::subclass::prelude::*;
+use gst_video::subclass::prelude::*;
 
-use atomic_refcell::AtomicRefCell;
 use std::i32;
 use std::sync::Mutex;
 
 use once_cell::sync::Lazy;
-use std::convert::TryInto;
 
 use super::super::hsvutils;
 
@@ -54,17 +55,10 @@ impl Default for Settings {
     }
 }
 
-// Stream-specific state, i.e. video format configuration
-struct State {
-    in_info: gst_video::VideoInfo,
-    out_info: gst_video::VideoInfo,
-}
-
 // Struct containing all the element data
 #[derive(Default)]
 pub struct HsvDetector {
     settings: Mutex<Settings>,
-    state: AtomicRefCell<Option<State>>,
 }
 
 static CAT: Lazy<gst::DebugCategory> = Lazy::new(|| {
@@ -79,7 +73,7 @@ static CAT: Lazy<gst::DebugCategory> = Lazy::new(|| {
 impl ObjectSubclass for HsvDetector {
     const NAME: &'static str = "HsvDetector";
     type Type = super::HsvDetector;
-    type ParentType = gst_base::BaseTransform;
+    type ParentType = gst_video::VideoFilter;
 }
 
 fn video_input_formats() -> Vec<glib::SendValue> {
@@ -173,7 +167,7 @@ impl ObjectImpl for HsvDetector {
     fn properties() -> &'static [glib::ParamSpec] {
         static PROPERTIES: Lazy<Vec<glib::ParamSpec>> = Lazy::new(|| {
             vec![
-                glib::ParamSpec::new_float(
+                glib::ParamSpecFloat::new(
                     "hue-ref",
                     "Hue reference",
                     "Hue reference in degrees",
@@ -182,7 +176,7 @@ impl ObjectImpl for HsvDetector {
                     DEFAULT_HUE_REF,
                     glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_PLAYING,
                 ),
-                glib::ParamSpec::new_float(
+                glib::ParamSpecFloat::new(
                     "hue-var",
                     "Hue variation",
                     "Allowed hue variation from the reference hue angle, in degrees",
@@ -191,7 +185,7 @@ impl ObjectImpl for HsvDetector {
                     DEFAULT_HUE_VAR,
                     glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_PLAYING,
                 ),
-                glib::ParamSpec::new_float(
+                glib::ParamSpecFloat::new(
                     "saturation-ref",
                     "Saturation reference",
                     "Reference saturation value",
@@ -200,7 +194,7 @@ impl ObjectImpl for HsvDetector {
                     DEFAULT_SATURATION_REF,
                     glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_PLAYING,
                 ),
-                glib::ParamSpec::new_float(
+                glib::ParamSpecFloat::new(
                     "saturation-var",
                     "Saturation variation",
                     "Allowed saturation variation from the reference value",
@@ -209,7 +203,7 @@ impl ObjectImpl for HsvDetector {
                     DEFAULT_SATURATION_VAR,
                     glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_PLAYING,
                 ),
-                glib::ParamSpec::new_float(
+                glib::ParamSpecFloat::new(
                     "value-ref",
                     "Value reference",
                     "Reference value value",
@@ -218,7 +212,7 @@ impl ObjectImpl for HsvDetector {
                     DEFAULT_VALUE_REF,
                     glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_PLAYING,
                 ),
-                glib::ParamSpec::new_float(
+                glib::ParamSpecFloat::new(
                     "value-var",
                     "Value variation",
                     "Allowed value variation from the reference value",
@@ -350,6 +344,8 @@ impl ObjectImpl for HsvDetector {
     }
 }
 
+impl GstObjectImpl for HsvDetector {}
+
 impl ElementImpl for HsvDetector {
     fn metadata() -> Option<&'static gst::subclass::ElementMetadata> {
         static ELEMENT_METADATA: Lazy<gst::subclass::ElementMetadata> = Lazy::new(|| {
@@ -366,21 +362,18 @@ impl ElementImpl for HsvDetector {
 
     fn pad_templates() -> &'static [gst::PadTemplate] {
         static PAD_TEMPLATES: Lazy<Vec<gst::PadTemplate>> = Lazy::new(|| {
-            let caps = gst::Caps::new_simple(
-                "video/x-raw",
-                &[
-                    ("format", &gst::List::from_owned(video_output_formats())),
-                    ("width", &gst::IntRange::<i32>::new(0, i32::MAX)),
-                    ("height", &gst::IntRange::<i32>::new(0, i32::MAX)),
-                    (
-                        "framerate",
-                        &gst::FractionRange::new(
-                            gst::Fraction::new(0, 1),
-                            gst::Fraction::new(i32::MAX, 1),
-                        ),
+            let caps = gst::Caps::builder("video/x-raw")
+                .field("format", gst::List::from(video_output_formats()))
+                .field("width", gst::IntRange::new(0, i32::MAX))
+                .field("height", gst::IntRange::new(0, i32::MAX))
+                .field(
+                    "framerate",
+                    gst::FractionRange::new(
+                        gst::Fraction::new(0, 1),
+                        gst::Fraction::new(i32::MAX, 1),
                     ),
-                ],
-            );
+                )
+                .build();
 
             let src_pad_template = gst::PadTemplate::new(
                 "src",
@@ -391,21 +384,18 @@ impl ElementImpl for HsvDetector {
             .unwrap();
 
             // sink pad capabilities
-            let caps = gst::Caps::new_simple(
-                "video/x-raw",
-                &[
-                    ("format", &gst::List::from_owned(video_input_formats())),
-                    ("width", &gst::IntRange::<i32>::new(0, i32::MAX)),
-                    ("height", &gst::IntRange::<i32>::new(0, i32::MAX)),
-                    (
-                        "framerate",
-                        &gst::FractionRange::new(
-                            gst::Fraction::new(0, 1),
-                            gst::Fraction::new(i32::MAX, 1),
-                        ),
+            let caps = gst::Caps::builder("video/x-raw")
+                .field("format", gst::List::from(video_input_formats()))
+                .field("width", gst::IntRange::new(0, i32::MAX))
+                .field("height", gst::IntRange::new(0, i32::MAX))
+                .field(
+                    "framerate",
+                    gst::FractionRange::new(
+                        gst::Fraction::new(0, 1),
+                        gst::Fraction::new(i32::MAX, 1),
                     ),
-                ],
-            );
+                )
+                .build();
 
             let sink_pad_template = gst::PadTemplate::new(
                 "sink",
@@ -438,11 +428,11 @@ impl BaseTransformImpl for HsvDetector {
         let mut other_caps = caps.clone();
         if direction == gst::PadDirection::Src {
             for s in other_caps.make_mut().iter_mut() {
-                s.set("format", &gst::List::from_owned(video_input_formats()));
+                s.set("format", gst::List::from(video_input_formats()));
             }
         } else {
             for s in other_caps.make_mut().iter_mut() {
-                s.set("format", &gst::List::from_owned(video_output_formats()));
+                s.set("format", gst::List::from(video_output_formats()));
             }
         };
 
@@ -463,90 +453,22 @@ impl BaseTransformImpl for HsvDetector {
             Some(other_caps)
         }
     }
+}
 
-    fn unit_size(&self, _element: &Self::Type, caps: &gst::Caps) -> Option<usize> {
-        gst_video::VideoInfo::from_caps(caps)
-            .map(|info| info.size())
-            .ok()
-    }
-
-    fn set_caps(
+impl VideoFilterImpl for HsvDetector {
+    fn transform_frame(
         &self,
-        element: &Self::Type,
-        incaps: &gst::Caps,
-        outcaps: &gst::Caps,
-    ) -> Result<(), gst::LoggableError> {
-        let in_info = match gst_video::VideoInfo::from_caps(incaps) {
-            Err(_) => return Err(gst::loggable_error!(CAT, "Failed to parse input caps")),
-            Ok(info) => info,
-        };
-        let out_info = match gst_video::VideoInfo::from_caps(outcaps) {
-            Err(_) => return Err(gst::loggable_error!(CAT, "Failed to parse output caps")),
-            Ok(info) => info,
-        };
-
-        gst_debug!(
-            CAT,
-            obj: element,
-            "Configured for caps {} to {}",
-            incaps,
-            outcaps
-        );
-
-        *self.state.borrow_mut() = Some(State { in_info, out_info });
-
-        Ok(())
-    }
-
-    fn stop(&self, element: &Self::Type) -> Result<(), gst::ErrorMessage> {
-        // Drop state
-        *self.state.borrow_mut() = None;
-
-        gst_info!(CAT, obj: element, "Stopped");
-
-        Ok(())
-    }
-
-    fn transform(
-        &self,
-        element: &Self::Type,
-        inbuf: &gst::Buffer,
-        outbuf: &mut gst::BufferRef,
+        _element: &Self::Type,
+        in_frame: &gst_video::VideoFrameRef<&gst::BufferRef>,
+        out_frame: &mut gst_video::VideoFrameRef<&mut gst::BufferRef>,
     ) -> Result<gst::FlowSuccess, gst::FlowError> {
-        let mut state_guard = self.state.borrow_mut();
-        let state = state_guard.as_mut().ok_or(gst::FlowError::NotNegotiated)?;
-
-        let in_frame =
-            gst_video::VideoFrameRef::from_buffer_ref_readable(inbuf.as_ref(), &state.in_info)
-                .map_err(|_| {
-                    gst::element_error!(
-                        element,
-                        gst::CoreError::Failed,
-                        ["Failed to map input buffer readable"]
-                    );
-                    gst::FlowError::Error
-                })?;
-
-        // And now map the output buffer writable, so we can fill it.
-        let mut out_frame =
-            gst_video::VideoFrameRef::from_buffer_ref_writable(outbuf, &state.out_info).map_err(
-                |_| {
-                    gst::element_error!(
-                        element,
-                        gst::CoreError::Failed,
-                        ["Failed to map output buffer writable"]
-                    );
-                    gst::FlowError::Error
-                },
-            )?;
-
-        match state.in_info.format() {
+        match in_frame.format() {
             gst_video::VideoFormat::Rgbx | gst_video::VideoFormat::Rgb => {
-                match state.out_info.format() {
+                match out_frame.format() {
                     gst_video::VideoFormat::Rgba => {
                         self.hsv_detect(
-                            &in_frame,
-                            &mut out_frame,
+                            in_frame,
+                            out_frame,
                             |in_p| {
                                 hsvutils::from_rgb(
                                     in_p[..3].try_into().expect("slice with incorrect length"),
@@ -560,8 +482,8 @@ impl BaseTransformImpl for HsvDetector {
                     }
                     gst_video::VideoFormat::Argb => {
                         self.hsv_detect(
-                            &in_frame,
-                            &mut out_frame,
+                            in_frame,
+                            out_frame,
                             |in_p| {
                                 hsvutils::from_rgb(
                                     in_p[..3].try_into().expect("slice with incorrect length"),
@@ -575,8 +497,8 @@ impl BaseTransformImpl for HsvDetector {
                     }
                     gst_video::VideoFormat::Bgra => {
                         self.hsv_detect(
-                            &in_frame,
-                            &mut out_frame,
+                            in_frame,
+                            out_frame,
                             |in_p| {
                                 hsvutils::from_rgb(
                                     in_p[..3].try_into().expect("slice with incorrect length"),
@@ -592,8 +514,8 @@ impl BaseTransformImpl for HsvDetector {
                     }
                     gst_video::VideoFormat::Abgr => {
                         self.hsv_detect(
-                            &in_frame,
-                            &mut out_frame,
+                            in_frame,
+                            out_frame,
                             |in_p| {
                                 hsvutils::from_rgb(
                                     in_p[..3].try_into().expect("slice with incorrect length"),
@@ -611,11 +533,11 @@ impl BaseTransformImpl for HsvDetector {
                 }
             }
             gst_video::VideoFormat::Xrgb => {
-                match state.out_info.format() {
+                match out_frame.format() {
                     gst_video::VideoFormat::Rgba => {
                         self.hsv_detect(
-                            &in_frame,
-                            &mut out_frame,
+                            in_frame,
+                            out_frame,
                             |in_p| {
                                 hsvutils::from_rgb(
                                     in_p[1..4].try_into().expect("slice with incorrect length"),
@@ -629,8 +551,8 @@ impl BaseTransformImpl for HsvDetector {
                     }
                     gst_video::VideoFormat::Argb => {
                         self.hsv_detect(
-                            &in_frame,
-                            &mut out_frame,
+                            in_frame,
+                            out_frame,
                             |in_p| {
                                 hsvutils::from_rgb(
                                     in_p[1..4].try_into().expect("slice with incorrect length"),
@@ -644,8 +566,8 @@ impl BaseTransformImpl for HsvDetector {
                     }
                     gst_video::VideoFormat::Bgra => {
                         self.hsv_detect(
-                            &in_frame,
-                            &mut out_frame,
+                            in_frame,
+                            out_frame,
                             |in_p| {
                                 hsvutils::from_rgb(
                                     in_p[1..4].try_into().expect("slice with incorrect length"),
@@ -661,8 +583,8 @@ impl BaseTransformImpl for HsvDetector {
                     }
                     gst_video::VideoFormat::Abgr => {
                         self.hsv_detect(
-                            &in_frame,
-                            &mut out_frame,
+                            in_frame,
+                            out_frame,
                             |in_p| {
                                 hsvutils::from_rgb(
                                     in_p[1..4].try_into().expect("slice with incorrect length"),
@@ -680,11 +602,11 @@ impl BaseTransformImpl for HsvDetector {
                 };
             }
             gst_video::VideoFormat::Bgrx | gst_video::VideoFormat::Bgr => {
-                match state.out_info.format() {
+                match out_frame.format() {
                     gst_video::VideoFormat::Rgba => {
                         self.hsv_detect(
-                            &in_frame,
-                            &mut out_frame,
+                            in_frame,
+                            out_frame,
                             |in_p| {
                                 hsvutils::from_bgr(
                                     in_p[..3].try_into().expect("slice with incorrect length"),
@@ -700,8 +622,8 @@ impl BaseTransformImpl for HsvDetector {
                     }
                     gst_video::VideoFormat::Argb => {
                         self.hsv_detect(
-                            &in_frame,
-                            &mut out_frame,
+                            in_frame,
+                            out_frame,
                             |in_p| {
                                 hsvutils::from_bgr(
                                     in_p[..3].try_into().expect("slice with incorrect length"),
@@ -717,8 +639,8 @@ impl BaseTransformImpl for HsvDetector {
                     }
                     gst_video::VideoFormat::Bgra => {
                         self.hsv_detect(
-                            &in_frame,
-                            &mut out_frame,
+                            in_frame,
+                            out_frame,
                             |in_p| {
                                 hsvutils::from_bgr(
                                     in_p[..3].try_into().expect("slice with incorrect length"),
@@ -732,8 +654,8 @@ impl BaseTransformImpl for HsvDetector {
                     }
                     gst_video::VideoFormat::Abgr => {
                         self.hsv_detect(
-                            &in_frame,
-                            &mut out_frame,
+                            in_frame,
+                            out_frame,
                             |in_p| {
                                 hsvutils::from_bgr(
                                     in_p[..3].try_into().expect("slice with incorrect length"),
@@ -748,11 +670,11 @@ impl BaseTransformImpl for HsvDetector {
                     _ => unreachable!(),
                 }
             }
-            gst_video::VideoFormat::Xbgr => match state.out_info.format() {
+            gst_video::VideoFormat::Xbgr => match out_frame.format() {
                 gst_video::VideoFormat::Rgba => {
                     self.hsv_detect(
-                        &in_frame,
-                        &mut out_frame,
+                        in_frame,
+                        out_frame,
                         |in_p| {
                             hsvutils::from_bgr(
                                 in_p[1..4].try_into().expect("slice with incorrect length"),
@@ -768,8 +690,8 @@ impl BaseTransformImpl for HsvDetector {
                 }
                 gst_video::VideoFormat::Argb => {
                     self.hsv_detect(
-                        &in_frame,
-                        &mut out_frame,
+                        in_frame,
+                        out_frame,
                         |in_p| {
                             hsvutils::from_bgr(
                                 in_p[1..4].try_into().expect("slice with incorrect length"),
@@ -785,8 +707,8 @@ impl BaseTransformImpl for HsvDetector {
                 }
                 gst_video::VideoFormat::Bgra => {
                     self.hsv_detect(
-                        &in_frame,
-                        &mut out_frame,
+                        in_frame,
+                        out_frame,
                         |in_p| {
                             hsvutils::from_bgr(
                                 in_p[1..4].try_into().expect("slice with incorrect length"),
@@ -800,8 +722,8 @@ impl BaseTransformImpl for HsvDetector {
                 }
                 gst_video::VideoFormat::Abgr => {
                     self.hsv_detect(
-                        &in_frame,
-                        &mut out_frame,
+                        in_frame,
+                        out_frame,
                         |in_p| {
                             hsvutils::from_bgr(
                                 in_p[1..4].try_into().expect("slice with incorrect length"),
